import logging
import gzip
import io
import time
import numpy as np
import time
from .util import encode_chromosome


def get_variant_type(vcf_line):

    l = vcf_line.split()
    if len(l[3]) == len(l[4]):
        return "SNP"
    elif "VT=SNP" in vcf_line:
        return "SNP"
    elif len(l[3]) > 1 and len(l[4]) > 1:
        return "SUBSTITUTION"
    elif len(l[3]) > len(l[4]) and len(l[4]) == 1:
        return "DELETION"
    elif len(l[3]) < len(l[4]) and len(l[3]) == 1:
        return "INSERTION"
    elif "VT=INDEL" in vcf_line:
        if len(l[3]) > len(l[4]):
            return "DELETION"
        else:
            return "INSERTION"
    else:
        return ""
        raise Exception("Unsupported variant type on line %s" % vcf_line)


class TruthRegions:
    def __init__(self, file_name):
        self.regions = []
        f = open(file_name)

        for line in f:
            l = line.split()
            start = int(l[1])
            end = int(l[2])

            self.regions.append((start, end))

    def is_inside_regions(self, position):
        for region in self.regions:
            if position >= region[0] and position < region[1]:
                return True
        return False


class VcfVariant:
    def __init__(self, chromosome, position, ref_sequence="", variant_sequence="", genotype=None, type="", vcf_line=None, vcf_line_number=None):
        self.chromosome = chromosome
        self.position = position
        self.ref_sequence = ref_sequence
        self.variant_sequence = variant_sequence
        self.genotype = genotype
        self.vcf_line = vcf_line
        if self.genotype == "1|0":
            self.genotype = "0|1"

        self.type = type
        self.vcf_line_number = vcf_line_number

        self._filter = "PASS"

    def get_genotype(self):
        return self.genotype.replace("|", "/").replace("1/0", "0/1").replace("./.", "0/0").replace(".", "0/0")

    def copy(self):
        return VcfVariant(self.chromosome, self.position, self.ref_sequence, self.variant_sequence, self.genotype, self.type, self.vcf_line, self.vcf_line_number)

    def set_filter_by_prob(self, prob_correct, criteria_for_pass=0.99):
        if prob_correct > criteria_for_pass:
            self._filter = "PASS"
        else:
            self._filter = "LowQUAL"

    def get_numeric_genotype(self):
        g = self.get_genotype()
        if g == "0/0":
            return 1
        elif g == "1/1":
            return 2
        elif g == "0/1":
            return 3
        else:
            raise Exception("Invalid genotype. Genotype is: %s" % g)


    def set_genotype(self, genotype, is_numeric=False):
        if is_numeric:
            assert genotype in [0, 1, 2, 3]
            if genotype == 1 or genotype == 0:
                self.genotype = "0/0"
            elif genotype == 2:
                self.genotype = "1/1"
            else:
                self.genotype = "0/1"
        else:
            assert genotype in ["0|0", "0/0", "0|1", "0/1", "1/1", "1|1"], "Invalid genotype %s" % genotype
            self.genotype = genotype

    def id(self):
        #return (self.chromosome, self.position)
        return (self.chromosome, self.position, self.ref_sequence.lower(), self.variant_sequence.lower())

    def get_vcf_line(self):
        chromosome = self.chromosome
        if chromosome == 23:
            chromosome = "X"
        elif chromosome == 24:
            chromosome = "Y"
        else:
            chromosome = str(chromosome)

        return "%s\t%d\t.\t%s\t%s\t.\t%s\t.\tGT\t%s\n"  % (chromosome, self.position, self.ref_sequence, self.variant_sequence, self._filter, self.genotype if self.genotype is not None else ".")

    def __str__(self):
        return "%s:%d %s/%s %s %s" % (self.chromosome, self.position, self.ref_sequence, self.variant_sequence, self.genotype, self.type)

    def length(self):
        if self.type == "SNP":
            return 1
        elif self.type == "DELETION":
            return len(self.get_deleted_sequence())
        elif self.type == "INSERTION" or self.type == "SUBSTITUTION":
            return len(self.variant_sequence) - 1
        else:
            raise Exception("Variant type %s not implemented" % self.type)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if self.id() == other.id():
            if self.genotype == other.genotype:
                return True
        else:
            logging.error("ID mismatch: %s != %s" % (self.id(), other.id()))

        return False

    def get_deleted_sequence(self):
        if self.type == "DELETION":
            return self.ref_sequence[1:]

        raise Exception("Variant %s is not a deletion" % self)

    def get_variant_sequence(self):
        if self.type == "DELETION":
            return ""
        elif self.type == "INSERTION" or self.type == "SUBSTITUTION":
            return self.variant_sequence[1:]
        else:
            return self.variant_sequence

    def get_inserted_sequence(self):
        # different from get_variant_sequence for deletions
        if self.type == "DELETION":
            inserted_sequence = self.ref_sequence[1:]
        elif self.type == "INSERTION":
            inserted_sequence = self.variant_sequence[1:]
        elif self.type == "SUBSTITUTION":
            inserted_sequence = self.variant_sequence[1:]
        else:
            raise Exception("Unsupported variant")

        return inserted_sequence

    def get_reference_sequence(self):
        if self.type == "INSERTION":
            return ""
        elif self.type == "INSERTION" or self.type == "SUBSTITUTION":
            return self.ref_sequence[1:]
        else:
            return self.ref_sequence

    def get_variant_allele_frequency(self):
        try:
            info_field = self.vcf_line.split("\t")[7]
            af = float(info_field.split("AF=")[1].split(";")[0])
        except IndexError:
            logging.error("Could not get allele frequency from vcf line. VCF needs to contain an info column with AF= some allele frequency")
            raise
        return af

    def get_reference_allele_frequency(self):
        return 1 - self.get_variant_allele_frequency()

    def get_individuals_and_numeric_haplotypes(self):
        # genotypes are 1, 2, 3 for homo ref, homo alt and hetero
        for individual_id, genotype_string in enumerate(self.vcf_line.split()[9:]):
            genotype_string = genotype_string.split(":")[0].replace("/", "|")
            if genotype_string == "0|0":
                numeric = (1, 1)
            elif genotype_string == "1|1":
                numeric = (2, 2)
            elif genotype_string == "0|1":
                numeric = (1, 2)
            elif genotype_string == "1|0":
                numeric = (2, 1)
            else:
                logging.error("Could not parse genotype string %s" % genotype_string)
                raise Exception("Unknown genotype")

            yield (individual_id, numeric)

    def _numeric_genotype(self, genotype_string):
        numeric = 0
        if genotype_string == "0|0":
            numeric = 1
        elif genotype_string == "1|1":
            numeric = 2
        elif genotype_string == "0|1" or genotype_string == "1|0":
            numeric = 3
        else:
            logging.error("Could not parse genotype string %s" % genotype_string)
            raise Exception("Unknown genotype")

        return numeric

    def _numeric_genotype2(self, genotype_string):
        numeric = 4
        if genotype_string == "0|0":
            numeric = 0
        elif genotype_string == "1|1":
            numeric = 2
        elif genotype_string == "0|1" or genotype_string == "1|0":
            numeric = 1
        elif genotype_string in [".", ".|.", "./."]:
            numberic = 4
        else:
            logging.error("Could not parse genotype string %s" % genotype_string)
            raise Exception("Unknown genotype")

        return numeric

    def _numeric_genotype3(self, genotype_string):
        numeric = 3
        if genotype_string == "0|0":
            numeric = 0
        elif genotype_string == "1|1":
            numeric = 2
        elif genotype_string == "0|1" or genotype_string == "1|0":
            numeric = 1
        else:
            logging.error("Could not parse genotype string %s" % genotype_string)
            raise Exception("Unknown genotype")

        return numeric

    def get_individuals_and_numeric_genotypes(self, encoding_version="1"):
        # genotypes are 1, 2, 3 for homo ref, homo alt and hetero
        for individual_id, genotype_string in enumerate(self.vcf_line.split()[9:]):
            genotype_string = genotype_string.split(":")[0].replace("/", "|")
            if encoding_version == "1":
                numeric = self._numeric_genotype(genotype_string)
            elif encoding_version == "2":
                numeric = self._numeric_genotype2(genotype_string)
            elif encoding_version == "3":
                numeric = self._numeric_genotype3(genotype_string)
            else:
                raise Exception("Invalid encoding version. Must be 1 or 2")
            yield (individual_id, numeric)

    @classmethod
    def from_vcf_line(cls, line, vcf_line_number=None, dont_encode_chromosome=True):
        l = line.split()
        chromosome = l[0]
        if not dont_encode_chromosome:
            assert False, "Encoding chromosome is not possible anymore"
            chromosome = int(encode_chromosome(l[0]))

        position = int(l[1])
        ref_sequence = l[3].lower().replace("n", "a")  # we only want actg
        variant_sequence = l[4].lower().replace("n", "a")

        if len(l) >= 10:
            genotype = l[9].split(":")[0].replace("/", "|")
        else:
            genotype = ""

        #logging.info("Genotype is %s. Line length is %d (%s) " % (genotype, len(l), l))

        return cls(chromosome, position, ref_sequence, variant_sequence, genotype, get_variant_type(line), line, vcf_line_number=vcf_line_number)

    def get_reference_position_before_variant(self):
        # Returns the position of the last base pair before the variant starts (will be end of node before variant)
        # For SNPS, the position is the actual SNP node
        # FOr deletions, the position is one basepair before the deleted sequence
        # For insertion, the position is one baseiapr before the inserted sequence
        # Subtract -1 in the end to make it 0-based
        if self.type == "SNP":
            return self.position - 1 - 1
        else:
            return self.position - 1

    def get_reference_position_after_variant(self):
        # Returns the next reference position after the variant is finished
        start = self.get_reference_position_before_variant()
        if self.type == "SNP":
            return start + 2
        elif self.type == "INSERTION":
            return start + 1
        elif self.type == "DELETION":
            return start + len(self.ref_sequence) - 1 + 1
        elif self.type == "SUBSTITUTION":
            return start + len(self.ref_sequence)
        else:
            raise Exception("Not able to get ref pos after variant of type %s" % self.type)


class VcfVariants:
    def __init__(self, variant_genotypes=[], skip_index=False, header_lines=""):
        self._header_lines = header_lines
        self.variant_genotypes = variant_genotypes
        self._index = {}
        self._position_index = {}
        if not skip_index:
            self.make_index()

    def __getitem__(self, item):
        return self.variant_genotypes[item]

    def get_header(self):
        return self._header_lines

    def n_individuals(self):
        return len(list(self.variant_genotypes[0].get_individuals_and_numeric_genotypes()))

    def n_variants(self):
        assert isinstance(self.variant_genotypes, list), "Only possible when variants is a list"
        return len(self.variant_genotypes)

    def get_variant_by_line_number(self, line_number):
        return self.variant_genotypes[line_number]

    def add_variants(self, variant_list):
        self.variant_genotypes.extend(variant_list)

    def get_chunks(self, chunk_size=5000, add_variants_to_list=None):
        logging.info("Getting variant chunks of size %d" % chunk_size)
        out = []
        prev_time = time.time()
        i = 0
        for variant_number, variant in enumerate(self.variant_genotypes):
            if variant_number % 100000 == 0:
                logging.info("%d variants read" % variant_number)
            out.append(variant)
            if add_variants_to_list is not None:
                add_variants_to_list.append(variant)
            i += 1
            if i >= chunk_size and chunk_size > 0:
                logging.info("Reading chunk of %d variants took %.5f sec" % (chunk_size, time.time()-prev_time))
                prev_time = time.time()
                yield out
                out = []
                i = 0
        logging.info("Reading chunk of %d variants took %.5f sec" % (chunk_size, time.time() - prev_time))
        yield out

    def make_index(self):
        logging.info("Making vcf index")
        for variant in self.variant_genotypes:
            self._index[variant.id()] = variant
        logging.info("Done making vcf index")

    def apply_on_sequence(self, sequence, chromosome, sequence_start, sequence_end):
        assert self._position_index is not None, "Must create position index first"

        variants = self.get_variants_in_region(chromosome, sequence_start+1, sequence_end)

        sequence = list(sequence)
        n_changed = 0
        for variant in variants:
            if variant.get_genotype() == "0/1" or variant.get_genotype() == "1/1":
                if variant.type == "SNP":
                    position = variant.position - sequence_start - 1
                    assert sequence[position].lower() == variant.ref_sequence.lower(), "Variant %s does not match sequence %s. Position %s. Sequence at position: %s" % (variant, sequence, position, sequence[position-1:position+2])
                    sequence[position] = variant.variant_sequence
                    n_changed += 1

        #logging.info("Sequence changed %d times" % n_changed)
        return ''.join(sequence)

    def make_position_index(self):
        logging.info("Finding unique chromosomes")
        unique_chromosomes = np.unique([variant.chromosome for variant in self])

        logging.info("Making index")
        self._position_index = {chromosome: [] for chromosome in unique_chromosomes}
        self._variants_by_chromosome = {chromosome: [] for chromosome in unique_chromosomes}

        for variant in self:
            self._position_index[variant.chromosome].append(variant.position)
            self._variants_by_chromosome[variant.chromosome].append(variant)

        logging.info("Converting to numpy")
        for chrom in unique_chromosomes:
            self._position_index[chrom] = np.array(self._position_index[chrom])
        """
        for chrom in unique_chromosomes:
            logging.info("Making for chromosome %s" % chrom)
            self._position_index[chrom] = np.array([variant.position for variant in self if variant.chromosome == chrom])
            self._variants_by_chromosome[chrom] = [variant for variant in self if variant.chromosome == chrom]
        """

    @staticmethod
    def translate_numeric_genotypes_to_literal(genotypes):
        numeric_to_literal = {0: "0/0", 1: "0/0", 2: "1/1", 3: "0/1"}
        return [numeric_to_literal[g] for g in genotypes]

    def has_variant_left_of_variant(self, variant, window=31):
        other = self.get_variants_in_region(variant.chromosome, variant.position - window, variant.position)
        assert variant.position not in [v.position for v in other], "Variants between %d and %d contain %s position: %s" % (variant.position-window, variant.position, variant, other)
        if len(other) > 0:
            return True
        return False

    def has_variant_right_of_variant(self, variant, window=31):
        other = self.get_variants_in_region(variant.chromosome, variant.position+1, variant.position+window)
        assert variant.position not in [v.position for v in other]
        if len(other) > 0:
            return True
        return False

    def get_variants_in_region(self, chromosome, start, end):
        #chromosome = int(chromosome)
        if chromosome not in self._variants_by_chromosome:
            #logging.info("Invalid chromosome %s of type %s. Possible chromosomes are %s" % (chromosome, type(chromosome), self._variants_by_chromosome.keys()))
            return []
        start_index = np.searchsorted(self._position_index[chromosome], start)
        end_index = np.searchsorted(self._position_index[chromosome], end)
        return self._variants_by_chromosome[chromosome][start_index:end_index]

    def has_variant(self, variant_genotype):
        if variant_genotype.id() in self._index:
            return True

        return False

    def has_variant_genotype(self, variant_genotype):
        if self.has_variant(variant_genotype) and self._index[variant_genotype.id()].genotype == variant_genotype.genotype:
            return True

        return False

    def get(self, variant):
        if variant.id() not in self._index:
            return None
        return self._index[variant.id()]

    def __getitem__(self, item):
        if item >= len(self.variant_genotypes):
            logging.error("Variant with id %d does not exist. There are only %d variants" % (item, len(self.variant_genotypes)))

        return self.variant_genotypes[item]

    def __len__(self):
        return len(self.variant_genotypes)

    def __iter__(self):
        return self.variant_genotypes.__iter__()

    def __next__(self):
        return self.variant_genotypes.__next__()

    def to_vcf_file(self, file_name, add_individual_to_header="DONOR", ignore_homo_ref=False, add_header_lines=None, sample_name_output="DONOR"):
        logging.info("Writing to file %s" % file_name)
        with open(file_name, "w") as f:
            #f.write(self._header_lines)
            for header_line in self._header_lines.split("\n")[0:-1]:  # last element is empty
                if header_line.startswith("#CHROM"):
                    if sample_name_output != "":
                        header_line += "\t" + sample_name_output
                    # first write additional header line, the chrom-line should be the last one
                    if add_header_lines is not None:
                        for additional_header_line in add_header_lines:
                            f.writelines([additional_header_line + "\n"])

                f.writelines([header_line + "\n"])


            for i, variant in enumerate(self):
                if i % 1000000 == 0:
                    logging.info("%d variants written to file" % i)

                if ignore_homo_ref and (variant.get_genotype() == "0/0" or variant.get_genotype() == ""):
                    continue

                f.writelines([variant.get_vcf_line()])


    @staticmethod
    def read_header_from_vcf(vcf_file_name):
        f = open(vcf_file_name)
        is_bgzipped = False
        if vcf_file_name.endswith(".gz"):
            logging.info("Assuming gzipped file")
            is_bgzipped = True
            gz = gzip.open(vcf_file_name)
            f = io.BufferedReader(gz, buffer_size=1000 * 1000 * 2)  # 2 GB buffer size?
            logging.info("Made gz file object")

        header_lines = ""
        for line in f:
            if is_bgzipped:
                line = line.decode("utf-8")

            if line.startswith("#"):
                header_lines += line
            else:
                break

        return header_lines

    @classmethod
    def from_vcf(cls, vcf_file_name, skip_index=False, limit_to_n_lines=None, make_generator=False, limit_to_chromosome=None, dont_encode_chromosomes=False):
        logging.info("Reading variants from file")
        variant_genotypes = []

        """
        if limit_to_chromosome is not None:
            if limit_to_chromosome == "X":
                limit_to_chromosome = "23"
            elif limit_to_chromosome == "Y":
                limit_to_chromosome = "24"
        """

        if limit_to_chromosome is not None:
            logging.info("Will only read variants from chromsome %s" % limit_to_chromosome)

        # Get header
        header_lines = VcfVariants.read_header_from_vcf(vcf_file_name)

        f = open(vcf_file_name)
        is_bgzipped = False
        if vcf_file_name.endswith(".gz"):
            logging.info("Assuming gzipped file")
            is_bgzipped = True
            gz = gzip.open(vcf_file_name)
            f = io.BufferedReader(gz, buffer_size=1000 * 1000 * 2)  # 2 GB buffer size?
            logging.info("Made gz file object")

        if make_generator:
            assert limit_to_chromosome is None, "Cannot both limit to chromosome and make generator (not implemented)"
            logging.info("Returning variant generator")
            if is_bgzipped:
                f = (line for line in f if not line.decode("utf-8").startswith("#"))
                return cls((VcfVariant.from_vcf_line(line.decode("utf-8"), vcf_line_number=i, dont_encode_chromosome=dont_encode_chromosomes) for i, line in enumerate(f) if not line.decode("utf-8").startswith("#")), skip_index=skip_index, header_lines=header_lines)
            else:
                f = (line for line in f if not line.startswith("#"))
                return cls((VcfVariant.from_vcf_line(line, vcf_line_number=i, dont_encode_chromosome=dont_encode_chromosomes) for i, line in enumerate(f)), skip_index=skip_index, header_lines=header_lines)

        n_variants_added = 0
        prev_time = time.time()
        variant_number = -1
        for i, line in enumerate(f):
            if is_bgzipped:
                line = line.decode("utf-8")

            if line.startswith("#"):
                continue

            variant_number += 1

            if i % 5000000 == 0:
                logging.info("Read %d variants from file (time: %.3f). %d variants added" % (i, time.time()-prev_time, n_variants_added))
                prev_time = time.time()

            if limit_to_n_lines is not None and i >= limit_to_n_lines:
                logging.warning("Limited to %d lines" % limit_to_n_lines)
                break


            variant = VcfVariant.from_vcf_line(line, vcf_line_number=variant_number, dont_encode_chromosome=dont_encode_chromosomes)
            if limit_to_chromosome is not None and variant.chromosome != limit_to_chromosome:
                if len(variant_genotypes) > 0:
                    logging.info("Stoppinng reading file since limiting to chromosome and now on new chromosome")
                    break
                else:
                    continue

            n_variants_added += 1
            variant_genotypes.append(variant)

        return cls(variant_genotypes, skip_index, header_lines)

    def __str__(self):
        return str(list(self))

    def __repr__(self):
        return self.__str__()

    def copy(self):
        return VcfVariants([v.copy() for v in self])

    def intersect(self, other_variants):
        new_variants = []
        for variant in self:
            if other_variants.has_variant(variant):
                new_variants.append(variant)

        logging.info("Variants before intersection: %d Variants after: %d" % (len(self.variant_genotypes), len(new_variants)))
        return VcfVariants(new_variants, skip_index=True, header_lines=self._header_lines)

    def print_variants_not_in_other(self, other):
        for variant in self:
            if not other.has_variant(variant):
                print("Only in set 1: %s" % variant)

    def print_diff(self, other):
        for variant in self:
            if not other.has_variant(variant):
                print("Only in set 1: %s" % variant)
            elif not other.has_variant_genotype(variant):
                print("Genotype mismatch between %s and %s" % (variant, other.get(variant)))

    def compute_similarity_to_other_variants_unsorted(self, other_variants):
        n_identical = 0
        for variant in self:
            if other_variants.has_variant(variant):
                n_identical += 1
        return n_identical / len(self)

    def compute_similarity_to_other_variants(self, other_variants):
        n_identical = 0
        for variant, other_variant in zip(self, other_variants):
            assert variant.position == other_variant.position
            if variant.genotype == other_variant.genotype:
                n_identical += 1

        return n_identical / len(self)


import itertools
import json
import logging
import re
import numpy as np
from .variants import VcfVariants
from collections import defaultdict
from .mutable_graph import MutableGraph
from dataclasses import dataclass
from npstructures import RaggedArray, HashTable
from shared_memory_wrapper import to_file, from_file
from .nplist import NpList
from .util import encode_chromosome


class VariantNotFoundException(Exception):
    pass

def remap_array(array, from_values, to_values):
    index = np.digitize(array.ravel(), from_values, right=True)
    try:
        res = to_values[index].reshape(array.shape)
    except IndexError:
        logging.error("Invalid byte values in sequence. Max value: %d" % np.max(array))
        raise
    return res

def convert_sequence_array_to_numeric(sequence):
    assert type(sequence) == np.ndarray
    sequences_as_byte_values = sequence.astype("|S1").view(np.int8)
    # from byte values
    from_values = np.array([65, 67, 71, 84, 97, 99, 103, 116], dtype=np.uint64)  # NB: Must be increasing
    # to internal base values for a, c, t, g
    to_values = np.array([0, 1, 2, 3, 0, 1, 2, 3], dtype=np.uint64)
    return remap_array(sequences_as_byte_values, from_values, to_values).astype(np.uint8)

numeric_to_letter_sequence = np.array(["A", "C", "G", "T"])

class Graph:
    def __init__(self, nodes,
                 sequences: RaggedArray,
                 edges: RaggedArray,
                 node_to_ref_offset=None, ref_offset_to_node=None, chromosome_start_nodes=None, allele_frequencies=None, linear_ref_nodes_index=None,
                 numeric_node_sequences=None, linear_ref_nodes_and_dummy_nodes_index=None):
        assert chromosome_start_nodes is None or isinstance(chromosome_start_nodes, dict), "Chromosome start nodes must be None or a dict from chromosome to start node"
        self.nodes = nodes
        self.edges = edges
        self.sequences = sequences

        self.node_to_ref_offset = node_to_ref_offset
        self.ref_offset_to_node = ref_offset_to_node
        self._linear_ref_nodes_cache = None
        self.chromosome_start_nodes = chromosome_start_nodes
        self.allele_frequencies = allele_frequencies
        self.numeric_node_sequences = numeric_node_sequences

        if self.nodes is not None and linear_ref_nodes_index is None:
            logging.info("Makng a linear ref lookup, since it is not provided")
            self.linear_ref_nodes_index = np.zeros(len(self.nodes), dtype=np.uint8)
            self.linear_ref_nodes_index[np.array(list(self.linear_ref_nodes()))] = 1
            logging.info("Done making linear ref nodes index")
        else:
            self.linear_ref_nodes_index = linear_ref_nodes_index
        
        if self.nodes is not None and linear_ref_nodes_and_dummy_nodes_index is None:
            self.linear_ref_nodes_and_dummy_nodes_index = None
            logging.info("Making linear ref nodes and dummy nodes index")
            self.make_linear_ref_node_and_ref_dummy_node_index()
        else:
            self.linear_ref_nodes_and_dummy_nodes_index = linear_ref_nodes_and_dummy_nodes_index


    def __eq__(self, other):
        t = np.all(self.nodes == other.nodes) and np.all(self.edges == other.edges) and np.all(self.sequences == other.sequences)
        if self.chromosome_start_nodes is not None:
            if other.chromosome_start_nodes is None:
                return False

            t &= np.all(self.chromosome_start_nodes == other.chromosome_start_nodes)

        return t

    def node_has_edges(self, node, edges):
        e = self.get_edges(node)
        if len(edges) != len(e):
            return False

        for edge in edges:
            if edge not in edges:
                return False

        return True

    def set_numeric_node_sequences(self):
        raise Exception("Unsupported")

    def get_all_nodes(self):
        node_to_n_edges = self.edges.shape[0]
        return np.union1d(np.where(self.nodes != 0)[0], np.where(node_to_n_edges > 0)[0])
        # This does not return nodes of size 0
        #return np.where(self.nodes != 0)[0]

    def __str__(self):
        #mutable_graph = self.to_mutable_graph()
        #return str(mutable_graph)
        return "Obgraph"

    def __repr__(self):
        return self.__str__()

    def to_mutable_graph(self):
        all_nodes = self.get_all_nodes()
        nodes = {node: self.get_node_size(node) for node in all_nodes}
        node_sequences = {node: self.get_node_sequence(node) for node in all_nodes}
        edges = {node: list(self.get_edges(node)) for node in all_nodes}
        return MutableGraph(nodes, node_sequences, edges, list(self.linear_ref_nodes()), self.node_to_ref_offset, self.ref_offset_to_node, self.chromosome_start_nodes, self.allele_frequencies)

    @classmethod
    def from_mutable_graph(cls, mutable_graph):
        return cls.from_dicts(mutable_graph.node_sequences, mutable_graph.edges, mutable_graph.linear_ref_nodes,
                              mutable_graph.chromosome_start_nodes)

    def get_node_size(self, node):
        return self.nodes[node]

    def get_node_allele_frequencies(self, nodes):
        if self.allele_frequencies is None:
            return np.ones(len(nodes))
        return self.allele_frequencies[nodes]

    def get_node_allele_frequency(self, node):
        if self.allele_frequencies is None:
            return 1.0
        return self.allele_frequencies[node]

    def get_node_subsequence(self, node, start, end):
        return self.get_node_sequence(node)[int(start):int(end)]

    def get_node_sequence(self, node):
        return ''.join(numeric_to_letter_sequence[self.sequences[node]])

    def max_node_id(self):
        return len(self.nodes)-1

    def get_numeric_node_sequence(self, node):
        return self.sequences[node]

        if self.get_node_size(node) == 0:
            return np.array([-1])

        index_pos = self.node_to_sequence_index[node]
        return self.numeric_node_sequences[index_pos:index_pos+self.get_node_size(node)]

    def get_numeric_base_sequence(self, node, offset):
        if self.get_node_size(node) == 0:
            return -1

        return self.get_numeric_node_sequence(int(node))[int(offset)]


    def get_numeric_node_sequences_by_chromosomes(self, nodes):
        assert nodes[0] == list(self.chromosome_start_nodes.values())[0], "First node must be first chromosome node"

        # yields one sequence per chromosome, assumes the nodes are sorted so that
        # nodes on same chromosome are together
        split_positions = list(np.where(np.in1d(nodes, list(self.chromosome_start_nodes.values())))[0])
        #split_positions = list(np.searchsorted(nodes, self.chromosome_start_nodes))
        split_positions.append(nodes[-1])

        for start, end in zip(split_positions[0:-1], split_positions[1:]):
            yield self.get_numeric_node_sequences(nodes[start:end])


    def get_numeric_node_sequences(self, nodes):
        return self.sequences[nodes].ravel()

    def get_nodes_sequences2(self, nodes):
        raise NotImplementedError("Use get_nodes_sequences instead")

    def get_nodes_sequence(self, nodes):
        # todo: This is slow, should be fixed
        sequences = []

        for node in nodes:
            sequences.extend(self.get_node_sequence(node))

        return ''.join(sequences)

    def get_node_offset_at_chromosome_and_chromosome_offset(self, chromosome, offset):
        #chromosome_position = chromosome - 1
        try:
            chromosome_start_node = self.chromosome_start_nodes[chromosome]
        except IndexError:
            chromosome_start_node = list(self.chromosome_start_nodes.values())[0]

        chromosome_offset = self.get_ref_offset_at_node(chromosome_start_node)
        real_offset = int(chromosome_offset + offset)
        node = self.get_node_at_chromosome_and_chromosome_offset(chromosome, offset)
        node_offset = self.get_ref_offset_at_node(node)
        return real_offset - node_offset

    def get_node_at_chromosome_and_chromosome_offset(self, chromosome, offset):
        #chromosome_position = chromosome - 1
        try:
            chromosome_start_node = self.chromosome_start_nodes[chromosome]
        except IndexError:
            if len(self.chromosome_start_nodes) == 1:
                #logging.info("Assuming chromosome start node is the only start node %d" % self.chromosome_start_nodes[0])
                chromosome_start_node = self.chromosome_start_nodes[0]
            else:
                logging.error(
                    "Could not find chromosome start position for chromosome %d. Chromosome start nodes are %s" % (
                        chromosome, self.chromosome_start_nodes))
                raise


        # Shift offset with chromosome start offset
        chromosome_offset = self.get_ref_offset_at_node(chromosome_start_node)
        real_offset = int(chromosome_offset + offset)
        return self.get_node_at_ref_offset(real_offset)


    def get_linear_ref_nodes_between_offsets(self, chromosome, start_offset, end_offset):
        # todo
        chromosome_position = chromosome - 1
        try:
            chromosome_start_node = self.chromosome_start_nodes[chromosome_position]
        except IndexError:
            logging.error(
                "Could not find chromosome start position for chromosome %d. Chromosome start nodes are %s" % (
                chromosome, self.chromosome_start_nodes))
            raise

        chromosome_offset = self.get_ref_offset_at_node(chromosome_start_node)
        return np.unique(self.ref_offset_to_node[int(chromosome_offset+start_offset):int(chromosome_offset+end_offset)])

    def get_edges(self, node):
        if node >= len(self.edges):
            return []

        return self.edges[node]

    def linear_ref_nodes(self):
        if self._linear_ref_nodes_cache is not None:
            return self._linear_ref_nodes_cache
        else:
            nodes = set(np.unique(self.ref_offset_to_node))
            self._linear_ref_nodes_cache = nodes
            return nodes

    def make_linear_ref_node_and_ref_dummy_node_index(self):
        linear_ref_nodes_and_dummy_nodes_index = np.zeros(len(self.linear_ref_nodes_index), dtype=np.uint8)
        for node in range(len(linear_ref_nodes_and_dummy_nodes_index)):
            if node % 100000 == 0:
                logging.info("%d nodes processed" % node)

            if self.is_linear_ref_node_or_linear_ref_dummy_node(node):
                linear_ref_nodes_and_dummy_nodes_index[node] = 1

        self.linear_ref_nodes_and_dummy_nodes_index = linear_ref_nodes_and_dummy_nodes_index

    def is_linear_ref_node_or_linear_ref_dummy_node(self, node):
        if self.linear_ref_nodes_and_dummy_nodes_index is not None:
            if self.linear_ref_nodes_and_dummy_nodes_index[node] == 1:
                return True
            else:
                return False

        if self.is_linear_ref_node(node):
            return True

        if self.get_node_size(node) == 0:
            # is a linear if it has a linear ref node as next node and the previous node on the linear ref from this linear ref node has an edge to node
            if len([next_node for next_node in self.get_edges(node) if self.is_linear_ref_node(next_node) and node in self.get_edges(self.get_node_at_ref_offset(int(self.get_ref_offset_at_node(next_node)-1)))]) > 0:
                return True

        return False

    def is_linear_ref_node(self, node):
        if self.linear_ref_nodes_index[node] != 0:
            return True
        return False

    def linear_ref_length(self):
        return len(self.ref_offset_to_node)

    def get_node_at_ref_offset(self, ref_offset):
        try:
            return self.ref_offset_to_node[ref_offset]
        except IndexError:
            logging.error("Invalid ref offset %s" % str(ref_offset))
            raise

    def get_ref_offset_at_node(self, node):
        return self.node_to_ref_offset[node]

    def get_chromosome_ref_offset_at_node(self, chromosome, node):
        try:
            chromosome_start_node = self.chromosome_start_nodes[chromosome]
        except IndexError:
            if len(self.chromosome_start_nodes) == 1:
                chromosome_start_node = list(self.chromosome_start_nodes.values())[0]
            else:
                logging.error(
                "Could not find chromosome start position for chromosome %d. Chromosome start nodes are %s" % (
                    chromosome, self.chromosome_start_nodes))
                raise
        chromosome_offset = self.get_ref_offset_at_node(chromosome_start_node)
        return self.get_ref_offset_at_node(node) - chromosome_offset

    def get_node_offset_at_ref_offset(self, ref_offset):
        node = self.get_node_at_ref_offset(ref_offset)
        offset_at_node = self.get_ref_offset_at_node(node)
        return ref_offset - offset_at_node

    def to_file(self, file_name):
        return to_file(self, file_name)

    @classmethod
    def from_file(cls, file_name):
        return from_file(file_name)

    def get_flat_graph(self):
        node_ids = list(np.where(self.nodes > 0)[0])
        node_sizes = list(self.nodes[node_ids])
        node_sequences = []
        for node in node_ids:
            node_sequences.append(list(self.get_node_sequence(node)))

        linear_ref_nodes = list(self.linear_ref_nodes())
        from_nodes = []
        to_nodes = []
        for from_node in node_ids:
            for to_node in self.get_edges(from_node):
                from_nodes.append(from_node)
                to_nodes.append(to_node)

        return node_ids, node_sequences, node_sizes, from_nodes, to_nodes, linear_ref_nodes, self.chromosome_start_nodes

    @classmethod
    def from_dicts(cls, node_sequences, edges, linear_ref_nodes, chromosome_start_nodes=None):
        assert linear_ref_nodes is not None
        logging.info("Making graph from dicts")
        nodes = np.sort(list(node_sequences.keys())).astype(np.uint32)
        node_ids = nodes
        max_node = nodes[-1]
        logging.info("Max node id is %d" % max_node)

        logging.info("Making sequences")
        node_sequences = [node_sequences[node] for node in nodes]
        node_sizes = [len(seq) for seq in node_sequences]
        node_sizes_array = np.zeros(nodes[-1]+1, dtype=np.uint32)
        node_sizes_array[nodes] = node_sizes
        nodes = node_sizes_array

        #from_nodes = []

        to_nodes = []
        n_edges = []

        to_nodes = np.concatenate([edges[n] for n in range(max_node+1) if n in edges]).astype(np.uint32)
        n_edges = np.array([len(edges[n]) if n in edges else 0 for n in range(max_node+1)], dtype=np.uint8)

        to_nodes_set = set(to_nodes)
        logging.info("Done preparing data from dicts")
        logging.info("There are %d node sequences" % len(node_sequences))

        # sequences
        sequence = np.array(list(''.join(node_sequences)))
        numeric_sequence = convert_sequence_array_to_numeric(sequence)
        sequences = RaggedArray(numeric_sequence, nodes, dtype=np.uint8)

        # linear ref index
        node_to_ref_offset = np.zeros(max_node + 1, np.uint64)
        ref_node_sizes = nodes[linear_ref_nodes]

        # print("Ref node sizes: %s"  % ref_node_sizes)
        ref_offsets = np.cumsum(ref_node_sizes)
        node_to_ref_offset[linear_ref_nodes[1:]] = ref_offsets[:-1]


        # Find last node to add to linear ref size
        last_node_in_graph = np.argmax(node_to_ref_offset)
        last_node_size = nodes[last_node_in_graph]
        logging.info("Last node in graph is %d with size %d" % (last_node_in_graph, last_node_size))

        ref_offset_to_node = np.zeros(int(np.max(node_to_ref_offset)) + last_node_size)
        index_positions = np.cumsum(ref_node_sizes)[:-1]
        # logging.info("INdex positions: %s" % index_positions)
        # logging.info("Linear ref nodes: %s" % linear_ref_nodes)
        # logging.info("Diff linear ref nodes: %s" % np.diff(linear_ref_nodes))
        ref_offset_to_node[index_positions] = np.diff(linear_ref_nodes)
        ref_offset_to_node[0] = linear_ref_nodes[0]
        # logging.info("Ref offset to node 1: %s" % ref_offset_to_node)
        ref_offset_to_node = np.cumsum(ref_offset_to_node, dtype=np.uint32)

        if chromosome_start_nodes is None:
            chromosome_start_nodes = [node for node in node_ids if node not in to_nodes_set]
            chromosome_start_nodes = {i+1: node for i, node in enumerate(chromosome_start_nodes)}
            logging.info("Seting chromosome start nodes to be %s" % chromosome_start_nodes)
        else:
            logging.info("Chromosome start nodes already set to %s" % chromosome_start_nodes)

        return Graph(node_sizes_array, sequences, RaggedArray(to_nodes, n_edges, dtype=np.uint32), node_to_ref_offset, ref_offset_to_node, chromosome_start_nodes)
        #return Graph.from_flat_nodes_and_edges(nodes, node_sequences, node_sizes, to_nodes, n_edges, np.array(linear_ref_nodes))

    @classmethod
    def from_flat_nodes_and_edges(cls, node_ids, node_sequences, node_sizes, to_nodes, n_edges, linear_ref_nodes, chromosome_start_nodes):
        raise Exception("This method has been removed.")

        logging.info("Chromosome start nodes are: %s" % chromosome_start_nodes)

        #assert node_ids[0] == 1, "Nodes must start with id 1. First id now is %d" % node_ids[0]

        max_node = np.max(node_ids)
        nodes = np.zeros(max_node+1, dtype=np.uint32)
        nodes[node_ids] = node_sizes

        # Node sequences
        node_sorting = np.argsort(node_ids)

        sequences_sorted = node_sequences[node_sorting]
        logging.info("Making node sequences")
        sequence = np.array(list(''.join(sequences_sorted)))
        numeric_sequence = convert_sequence_array_to_numeric(sequence)
        sequences = RaggedArray(numeric_sequence, nodes, dtype=np.uint8)





        logging.info("Sorting nodes")
        #sorting = np.argsort(from_nodes)
        #from_nodes = from_nodes[sorting]
        #to_nodes = to_nodes[sorting]

        edges = RaggedArray(to_nodes, n_edges, dtype=np.uint32)

        """
        logging.info("Making node index")
        diffs = np.ediff1d(from_nodes, to_begin=1)
        positions_of_unique_nodes = np.nonzero(diffs)[0]
        unique_nodes = from_nodes[positions_of_unique_nodes]

        logging.info("Making node to edge index")
        node_to_edge_index = np.zeros(max_node+1, dtype=np.uint32)
        node_to_n_edges = np.zeros(max_node+1, dtype=np.uint8)
        node_to_edge_index[unique_nodes] = positions_of_unique_nodes
        n_edges_numbers = np.ediff1d(positions_of_unique_nodes, to_end=len(from_nodes)-positions_of_unique_nodes[-1])
        node_to_n_edges[unique_nodes] = n_edges_numbers
        """

        #logging.info("Finding ref offsets")
        node_to_ref_offset = np.zeros(max_node+1, np.uint64)
        try:
            ref_node_sizes = nodes[linear_ref_nodes]
        except IndexError:
            logging.error("Problem with linear ref nodes, which are %s" % linear_ref_nodes)
            print(type(linear_ref_nodes))
            raise
        #print("Ref node sizes: %s"  % ref_node_sizes)
        ref_offsets = np.cumsum(ref_node_sizes)
        #print("Ref offsets: %s"  % ref_offsets)
        node_to_ref_offset[linear_ref_nodes[1:]] = ref_offsets[:-1]

        # Find last node to add to linear ref size
        last_node_in_graph = np.argmax(node_to_ref_offset)
        last_node_size = nodes[last_node_in_graph]
        #logging.info("Last node in graph is %d with size %d" % (last_node_in_graph, last_node_size))

        ref_offset_to_node = np.zeros(int(np.max(node_to_ref_offset)) + last_node_size)
        index_positions = np.cumsum(ref_node_sizes)[:-1]
        #logging.info("INdex positions: %s" % index_positions)
        #logging.info("Linear ref nodes: %s" % linear_ref_nodes)
        #logging.info("Diff linear ref nodes: %s" % np.diff(linear_ref_nodes))
        ref_offset_to_node[index_positions] = np.diff(linear_ref_nodes)
        ref_offset_to_node[0] = linear_ref_nodes[0]
        #logging.info("Ref offset to node 1: %s" % ref_offset_to_node)
        ref_offset_to_node = np.cumsum(ref_offset_to_node, dtype=np.uint32)

        return cls(nodes, sequences, edges, node_to_ref_offset, ref_offset_to_node, chromosome_start_nodes)

    @classmethod
    def from_vg_json_files(cls, file_names):

        node_sequences = []
        node_ids = []
        node_sizes = []
        edges_from = []
        edges_to = []
        linear_ref_nodes = []
        chromosomes = []
        chromosome_start_nodes = []
        n_nodes_added = 0
        n_edges_added = 0
        chromosome_offset = 0
        for file_name in file_names:
            # Hackish (send in chromosome on comand line later):
            path_found = False
            file = open(file_name)
            for line in file:
                json_object = json.loads(line)
                if "node" in json_object:
                    for node in json_object["node"]:
                        id = int(node["id"])
                        node_ids.append(id)
                        node_sizes.append(len(node["sequence"]))
                        node_sequence = list(re.sub(r'[^acgtn]', "n", node["sequence"].lower()))
                        node_sequences.append(node_sequence)
                        n_nodes_added += 1
                        if n_nodes_added % 100000 == 0:
                            logging.info("%d nodes added" % n_nodes_added)

                if "edge" in json_object:
                    for edge in json_object["edge"]:
                        from_node = int(edge["from"])
                        to_node = int(edge["to"])

                        edges_from.append(from_node)
                        edges_to.append(to_node)
                        n_edges_added += 1
                        if n_edges_added % 100000 == 0:
                            logging.info("%d edges added" % n_edges_added)

                if "path" in json_object:
                    for path in json_object["path"]:
                        assert not path_found, "Found multiple paths, not sure which is the reference path. Path now:" % path["name"]
                        logging.info("Found path %s, assuming this is the reference path" % path["name"])
                        nodes_in_path = [mapping["position"]["node_id"] for mapping in path["mapping"]]
                        linear_ref_nodes.extend(nodes_in_path)
                        chromosome_start_nodes.append(nodes_in_path[0])
                        path_found = True

                logging.info("Chromosome start nodes are: %s" % chromosome_start_nodes)

        node_ids = np.array(node_ids)
        node_sizes = np.array(node_sizes)
        edges_from = np.array(edges_from)
        edges_to = np.array(edges_to)
        linear_ref_nodes = np.array(linear_ref_nodes, dtype=np.uint32)
        chromosome_start_nodes = np.array(chromosome_start_nodes, dtype=np.uint32)
        logging.info("Chromosome start nodes are: %s" % chromosome_start_nodes)

        return cls.from_flat_nodes_and_edges(node_ids, node_sequences, node_sizes, edges_from, edges_to, linear_ref_nodes, chromosome_start_nodes)


    def get_snp_nodes(self, ref_offset, reference_bases, variant_bases, chromosome=1):
        node = self.get_node_at_chromosome_and_chromosome_offset(chromosome, ref_offset)
        node_offset = self.get_node_offset_at_chromosome_and_chromosome_offset(chromosome, ref_offset)

        if node_offset != 0:
            raise VariantNotFoundException("Node offset is %d at ref_offset %d. Variant bases: %s. Chromosome %d" % (node_offset, ref_offset, variant_bases, chromosome))

        prev_node = self.get_node_at_chromosome_and_chromosome_offset(chromosome, ref_offset - 1)

        _, possible_snp_nodes = self.find_nodes_from_node_that_matches_sequence(prev_node, variant_bases, [], [])

        if len(possible_snp_nodes) == 0:
            raise VariantNotFoundException("Did not find any possible snp nodes for variant at ref offset %s with variant sequence %s" % (ref_offset, variant_bases))

        next_ref_pos = ref_offset + len(reference_bases)
        next_ref_node = self.get_node_at_chromosome_and_chromosome_offset(chromosome, next_ref_pos)

        # attempt at simpler approach. Should only be one non-linear-ref-node that matches sequence?
        """
        possible_snp_nodes = [n[-1] for n in possible_snp_nodes if not self.is_linear_ref_node_or_linear_ref_dummy_node(n[-1])]
        # only keep nodes 

        if len(possible_snp_nodes) != 1:
            logging.info("Multiple or no nodes (%s) nodes with sequence %s from node %d" % (possible_snp_nodes, variant_bases, prev_node))
            logging.info(str([(node, self.get_node_sequence(node)) for node in possible_snp_nodes]))
            logging.info("ref node is %d" % node)
            logging.info("Ref offset is %d" % ref_offset)
            raise Exception("Error")

        return node, possible_snp_nodes[0]
        """

        for possible_snp_node in possible_snp_nodes:
            # Not true if deletion before snp
            #assert len(possible_snp_node) == 1, "SNP nodes should only be one node"
            if len([n for n in possible_snp_node if self.get_node_size(n) > 0]) != 1:
                logging.error("There should only be one non-empty SNP node")
                logging.info("Found nodes: %s" % (possible_snp_node))
                logging.info("prev node is %s, variant bases is %s" % (prev_node, variant_bases))
                raise Exception("error")

            potential_next = possible_snp_node[-1]  # Get last node, this will be the snp node if there are more nodes
            assert self.get_node_size(potential_next) > 0
            # Also require that this node shares next node with our ref snp node, if not this could be a false match against an indel node
            # this assumption does not work for subsitutions
            # why does this not work for substitutions??
            next_from_snp_node = self.get_edges(node)
            if len([n for n in next_from_snp_node if n in self.get_edges(potential_next)]) > 0:
                return node, potential_next

            # logging.info("Node after deletion: %d" % next_ref_node)
            if self.get_node_offset_at_chromosome_and_chromosome_offset(chromosome, next_ref_pos) != 0:
                raise VariantNotFoundException("Could not find next ref node")

            #logging.info("Next ref node: %d. Edges from potential: %s" % (next_ref_node, str(self.get_edges(potential_next))))
            if next_ref_node in self.get_edges(potential_next):
                return node, potential_next

        logging.error("Could not parse substitution at offset %d with bases %s. Next ref pos is %d. Next ref node is %d" % (ref_offset, variant_bases, next_ref_pos, next_ref_node))
        logging.info("Reference node for snp/subsitution is %d" % node)
        logging.error("Possible nodes are: %s" % possible_snp_nodes)
        raise VariantNotFoundException("Error")

    def get_deletion_nodes(self, ref_offset, deletion_length, deleted_sequence, chromosome=1):
        ref_offset += 1
        # Node is the reference node that is deleted (the first one if there are multiple)
        node = self.get_node_at_chromosome_and_chromosome_offset(chromosome, ref_offset)
        node_offset = self.get_node_offset_at_chromosome_and_chromosome_offset(chromosome, ref_offset)
        #logging.info("Processing deletion at ref pos %d with size %d. Node inside deletion: %d" % (ref_offset, deletion_length, node))

        if node_offset != 0:
            raise VariantNotFoundException("Node offset is %d, not 0" %  (node_offset))

        prev_node = self.get_node_at_chromosome_and_chromosome_offset(chromosome, ref_offset - 1)
        #logging.info("Node before deletion: %d" % prev_node)

        # Find next reference node with offset corresponding to the number of deleted base pairs
        next_ref_pos = ref_offset + deletion_length
        next_ref_node = self.get_node_at_chromosome_and_chromosome_offset(chromosome, ref_offset + deletion_length)
        #logging.info("Node after deletion: %d" % next_ref_node)
        if self.get_node_offset_at_chromosome_and_chromosome_offset(chromosome, next_ref_pos) != 0:
            #logging.error("Offset %d is not at beginning of node" % next_ref_pos)
            #logging.error("Node at %d: %d" % (next_ref_pos, next_ref_node))
            #logging.error("Ref length in deletion: %s" % deletion_length)
            #logging.info("Ref pos beginning of deletion: %d" % ref_offset)

            raise VariantNotFoundException("Deletion at ref offset %d not in graph" % ref_offset)

        # Find an empty node between prev_node and next_ref_node
        # This solution should handle cases where the reference has multiple nodes inside deleted segment (e.g. a SNP inside a deleted sequence)
        # This is not true anymore, since a deltion can be followed by an insertion, leading to two or more dummy nodes before reaching the linear ref again
        #deletion_nodes = [node for node in self.get_edges(prev_node) if next_ref_node in self.get_edges(node) and self.get_node_size(node) == 0]
        deletion_nodes = [node for node in self.get_edges(prev_node) if self.get_node_size(node) == 0]
        #debug = [(node, self.get_edges(node), self.get_node_size(node)) for node in self.get_edges(prev_node)]
        #logging.info("%s" % debug)




        if len(deletion_nodes) != 1:

            # There can be more than one deletion node if there is an insertion right after a deletion
            if len(deletion_nodes) > 1:
                # If there are more than 1 deletion node, we find all paths that matches the deleted sequence, and try to find the deletion node that starts before and ends after this path
                one_path, possible_deleted_paths = self.find_nodes_from_node_that_matches_sequence(prev_node, deleted_sequence, [], [])
                assert len(one_path) >  0 and len(possible_deleted_paths) > 0, "Found no paths from node %d that matches the deleted sequence %s" % (prev_node, deleted_sequence)
                deletion_nodes_orig = deletion_nodes.copy()

                #print(possible_deleted_paths)
                deletion_nodes = [n for n in deletion_nodes if len([path for path in possible_deleted_paths if len([edge_from_deletion_node for edge_from_deletion_node in self.get_edges(n) if edge_from_deletion_node in self.get_edges(path[-1])]) > 0]) > 0]

                if len(deletion_nodes) == 0:
                    logging.warning("Error on deletion at pos %s. Node: %d. Deletion nodes: %s. Orig deletion nodes: %s. Edges going out: %s" % (ref_offset, node, deletion_nodes, deletion_nodes_orig, self.get_edges(node)))
                    logging.error("Could not find any nodes among possible deletion nodes %s that wrapped around any of the paths %s" % (deletion_nodes_orig, possible_deleted_paths))
                    raise VariantNotFoundException("Deletion not found")
                elif len(deletion_nodes) > 1:
                    logging.warning("Deletion at ref offset %s has multiple empty nodes going out from %d: %s. They all wrap around a deleted path in %s" % (ref_offset, prev_node, deletion_nodes, possible_deleted_paths))
                    if deletion_nodes[0] not in self.get_edges(node):
                        logging.warning("Error on deletion. Deletion nodes: %s. Edges going out: %s" % (deletion_nodes, self.get_edges(node)))
                        raise VariantNotFoundException("Deletion not found")
                    # Assert that there is an empty node between this deletion node and next ref node
                    # This assertion won't always hold, e.g. if there are two insertions after a deltion, so just skip it
                    #assert len([n for n in self.get_edges(deletion_nodes[0]) if next_ref_node in self.get_edges(n) and self.get_node_size(n) == 0]), "There is more than one deletion node, and the next node
            else:
                #logging.warning("Could not find deletion at ref offset %d in graph" % ref_offset)
                raise VariantNotFoundException("There should be only one deletion node between %d and %d for variant at ref pos %d. There are %d. Edges out from %d are: %s. Edges out from %d are %s" % (prev_node, next_ref_node, ref_offset, len(deletion_nodes), prev_node, self.get_edges(prev_node), node, self.get_edges(node)))

        deletion_node = deletion_nodes[0]
        # Not true anymore, e.g. if there is a snp right after a deletion
        #assert len(self.get_edges(deletion_node)) == 1, "Deletion node %d has not exactly 1 edge. Variant at ref offset %d. Edges are %s" % (deletion_node, ref_offset, self.get_edges(deletion_node))
        return (node, deletion_node)

    def get_insertion_nodes(self, variant, chromosome=1):
        ref_offset = variant.position
        #logging.info("Ref offset: %d" % ref_offset)
        # Find node right before insertion node
        node = self.get_node_at_chromosome_and_chromosome_offset(chromosome, ref_offset-1)
        #logging.info("Node at ref offset: %d" % node)
        node_offset = self.get_node_offset_at_chromosome_and_chromosome_offset(chromosome, ref_offset-1)
        node_size = self.get_node_size(node)
        insertion_length = len(variant.variant_sequence) - 1
        insertion_sequence = variant.variant_sequence[1:]  # First base is not included

        if node_offset != node_size - 1:
            raise VariantNotFoundException("Node offset %d is not at end of node %d which has size %d. Insertion not found in graph." % (node_offset, node, node_size))

        # Find out which next node matches the insertion
        # NB! There might be more than one node with the same sequence going out before the variant, since a SNP can have same sequence as the insertion
        next_ref_node = self.get_node_at_chromosome_and_chromosome_offset(chromosome, ref_offset)
        variant_node = None
        for potential_next in self.get_edges(node):
            #if potential_next in self.linear_ref_nodes():
            if self.is_linear_ref_node(potential_next):
                continue  # Next should not be in linear ref

            #print("Processing insertion at ref offset %d with base %s" % (ref_offset, base))
            #print("  Node %d with offset %d. Node size: %d" % (node, node_offset, self.blocks[node].length()))

            if self.get_node_size(potential_next) == 0:
                continue  # This is an empty deletion node

            #next_bases = self.get_node_sequence(potential_next)[0:insertion_length].upper()
            next_bases = self.get_node_subsequence(potential_next, 0, insertion_length).upper()
            if next_bases == insertion_sequence.upper():
                if next_ref_node in self.get_edges(potential_next):
                    variant_node = potential_next
                    break

        if variant_node is None:
            raise VariantNotFoundException("Could not find insertion node. No nodes going out from %d has sequence %s" % (node, variant.variant_sequence))

        # Find linear reference node (empty dummy node)
        assert next_ref_node in self.get_edges(variant_node), "Failed parsing insertion %s. Found %d as next ref node after variant node, but there is no edge from variantnode %d to %d" % (variant, next_ref_node, variant_node, next_ref_node)
        # If there are multiple insertions, one could find the correct dummy node by choosing the one that goes to next node with lowest ref pos
        dummy_nodes = [node for node in self.get_edges(node) if next_ref_node in self.get_edges(node) and self.get_node_size(node) == 0]
        if len(dummy_nodes) != 1:
            logging.error("Edges from %d are %s" % (node, self.get_edges(node)))
            logging.error("Error when parsing insertion %s. There are not exactly 1 insertion node between %d and %d. Nodes of length 0 between are %s" % (variant, node, next_ref_node, dummy_nodes))
            logging.error(f"Node sizes from {node} are " + str([self.get_node_size(n) for n in self.get_edges(node)]))
            raise Exception("")

        insertion_node = dummy_nodes[0]
        assert next_ref_node in self.get_edges(insertion_node), "Failed parsing insertion %s. Found %d as next ref node after dummy node, but there is no edge from dumy node %d to %d" % (variant, next_ref_node, insertion_node, next_ref_node)

        return (insertion_node, variant_node)

    def get_variant_nodes(self, variant):
        if variant.type == "SNP":
            return self.get_snp_nodes(variant.position-1, variant.get_reference_sequence(), variant.variant_sequence, variant.chromosome)
        elif variant.type == "SUBSTITUTION":
            return self.get_snp_nodes(variant.position, variant.get_reference_sequence(), variant.get_variant_sequence(), variant.chromosome)
        elif variant.type == "DELETION":
            return self.get_deletion_nodes(variant.position-1, len(variant.ref_sequence)-1, variant.get_deleted_sequence(), variant.chromosome)
        elif variant.type == "INSERTION":
            return self.get_insertion_nodes(variant, variant.chromosome)

        raise Exception("Invalid variant %s. Has no type set." % variant)

    def set_allele_frequencies_from_variants(self, variants, use_chromosome=None):
        frequencies = np.zeros(len(self.nodes), dtype=np.float16) + 1  # Set all to 1.0 initially

        for i, variant in enumerate(variants):
            if use_chromosome is not None:
                variant.chromosome = use_chromosome

            if i % 1000000 == 0:
                logging.info("%d variants processed" % i)
            try:
                reference_node, variant_node = self.get_variant_nodes(variant)
                reference_af = variant.get_reference_allele_frequency()
            except VariantNotFoundException:
                logging.info("Variant %s not found in graph" % variant)
                continue
            variant_af = variant.get_variant_allele_frequency()
            frequencies[reference_node] = reference_af
            frequencies[variant_node] = variant_af

        self.allele_frequencies = frequencies

    def get_variant_nodes_in_region(self, chromosome, linear_ref_start, linear_ref_end):
        nodes = set(self.get_node_at_chromosome_and_chromosome_offset(chromosome, pos) for pos in range(linear_ref_start, linear_ref_end))
        return [self.get_edges(node) for node in nodes if len(self.get_edges(node)) == 2]

    def get_first_node(self):
        # chromosome start nodes should always be set
        #return self.chromosome_start_nodes[0]

        for candidate in [0, 1]:
            if len(self.get_edges(candidate)) > 0:
                return candidate

        raise Exception("Couldn't find first graph node")

    def get_haplotype_node_paths_for_haplotypes(self, variants, limit_to_n_haplotypes=10):
        # First find all variant nodes that the haplotype has
        haplotypes = list(range(0, limit_to_n_haplotypes))
        variant_nodes_in_haplotype = defaultdict(set)
        for i, variant in enumerate(variants):
            if i % 1000 == 0:
                logging.info("%d variants processed" % i)
            reference_node, variant_node = self.get_variant_nodes(variant)

            genotypes = variant.vcf_line.split()[9:]
            for haplotype in haplotypes:
                individual_number = haplotype // 2
                haplotype_number = haplotype - individual_number * 2
                haplotype_string = genotypes[individual_number].replace("/", "|").split("|")[haplotype_number]
                if haplotype_string == "1":
                    # Follows the variant, add variant node here
                    variant_nodes_in_haplotype[haplotype].add(variant_node)
                else:
                    variant_nodes_in_haplotype[haplotype].add(reference_node)

        # Iterate graph
        nodes = np.zeros((len(haplotypes), len(self.nodes)), dtype=np.uint32)
        for haplotype in haplotypes:
            current_node = self.get_first_node()
            i = 0
            while True:
                #nodes[haplotype].append(current_node)
                nodes[haplotype, i] = current_node

                next_nodes = self.get_edges(current_node)
                if len(next_nodes) == 0:
                    break

                next_node = None
                if len(next_nodes) == 1:
                    next_node = next_nodes[0]
                else:
                    for potential_next in next_nodes:
                        if potential_next in variant_nodes_in_haplotype[haplotype]:
                            next_node = potential_next

                assert next_node is not None

                current_node = next_node
                i += 1

            nodes[haplotype, i] = current_node

        return nodes

    def get_reverse_edges_hashtable(self):
        from_nodes = NpList(dtype=np.uint32)
        to_nodes = NpList(dtype=np.uint32)

        for node in self.get_all_nodes():
            for edge in self.get_edges(node):
                from_nodes.append(edge)
                to_nodes.append(node)

        return HashTable(from_nodes.get_nparray(), to_nodes.get_nparray())

    def get_reverse_edges_dict(self):
        reverse_edges = defaultdict(list)
        for node in self.get_all_nodes():
            for edge in self.get_edges(node):
                reverse_edges[edge].append(node)

        return reverse_edges

    def convert_chromosome_ref_offset_to_graph_ref_offset(self, chromosome_offset, chromosome):
        # Add the graph ref offset at this chromosome to the chromosome offset
        #chromosome_position = chromosome - 1
        return int(self.node_to_ref_offset[self.chromosome_start_nodes[chromosome]] + chromosome_offset)


    def find_nodes_from_node_that_matches_sequence(self, from_node, sequence, nodes_found, all_paths_found):
        # hack to fix issue with "n"
        sequence = sequence.replace("n", "a").replace("N", "A")

        #logging.info("== Searching from node %d with sequence %s. Variant type %s. Nodes found: %s. All paths found: %s" % (from_node, sequence, variant_type, nodes_found, all_paths_found))
        if sequence == "":
            # All of the sequence is used successfully, return
            all_paths_found.append(nodes_found)
            #logging.info("   RETURNING. All paths found: %s" % all_paths_found)
            return nodes_found, all_paths_found

        next_nodes = self.get_edges(from_node)
        result = (nodes_found, all_paths_found)
        for possible_next in next_nodes:

            #print("Checking next node %d" % possible_next)

            node_size = self.get_node_size(possible_next)
            if node_size == 0 or self.get_node_sequence(possible_next).lower() == sequence[0:node_size].lower():
                # This node is a match, we can continue searching
                new_sequence = sequence[node_size:]
                new_nodes_found = nodes_found.copy()
                new_nodes_found.append(possible_next)

                #print("   Matching sequence. New sequence is now %s" % new_sequence)
                result = self.find_nodes_from_node_that_matches_sequence(possible_next, new_sequence, new_nodes_found, all_paths_found)
                if not result:
                    continue
            else:
                #print("   No sequence match")
                continue

        return result




from .variants import VcfVariants
# Modifies a graph so that deletions/insertions has parallel dummy nodes
import logging
from itertools import product
from .graph import Graph, VariantNotFoundException


class DummyNodeAdder:
    def __init__(self, graph,  variants):
        self.graph = graph
        self.variants = variants
        self.mutable_graph = None
        self.current_new_node_id = len(self.graph.nodes) + 1
        self._n_variants_failed = 0
        self._n_variants_fixed = 0
        self._old_edges_to_new_node_mapping = {}

    def create_new_graph_with_dummy_nodes(self, use_mutable_graph=None):
        if use_mutable_graph is None:
            logging.info("Creating a mutable graph that can be changed")
            self.mutable_graph = self.graph.to_mutable_graph()
        else:
            self.mutable_graph = use_mutable_graph

        logging.info("Adding dummy nodes")
        for i, variant in enumerate(self.variants):
            if i % 10000 == 0:
                logging.info("%d variants processed" % i)

            if variant.type == "SNP" or variant.type == "SUBSTITUTION":
                continue

            try:
                self._add_dummy_edges_around_indel(variant)
            except VariantNotFoundException as e:
                logging.error("Could not find variant: %s" % str(e))
                self._n_variants_failed += 1
                continue
            self._n_variants_fixed += 1

        logging.info("Creating a new immutable graph from the mutable graph")
        #print(self.mutable_graph.get_all_nodes())
        #print(self.mutable_graph.node_sequences)
        logging.info("%d variants were not found in graph" % self._n_variants_failed)
        logging.info("%d variants were found and dummy nodes were added for these" % self._n_variants_fixed)
        return Graph.from_mutable_graph(self.mutable_graph)

    def get_edge_mapping(self):
        return self._old_edges_to_new_node_mapping

    def get_nodes_for_inserted_sequence_at_ref_pos(self, variant):
        inserted_sequence = variant.get_inserted_sequence()
        node_before_inserted_nodes = self.graph.get_node_at_chromosome_and_chromosome_offset(variant.chromosome, variant.position-1)
        inserted_nodes = self.mutable_graph.find_nodes_from_node_that_matches_sequence(node_before_inserted_nodes, inserted_sequence, variant.type, [], [])
        if len(inserted_nodes) == 0:
            raise VariantNotFoundException("Could not find inserted nodes for sequence %s, variant %s. Node before is %d" % (inserted_sequence, variant, node_before_inserted_nodes))

        return inserted_nodes

    def _add_dummy_edges_around_indel(self, variant):
        inserted_sequence = variant.get_inserted_sequence()
        _, possible_inserted_node_paths = self.get_nodes_for_inserted_sequence_at_ref_pos(variant)

        #logging.info("===========")
        #logging.info("Variant %s. Inserted nodes: %s" % (variant, possible_inserted_node_paths))
        assert len(possible_inserted_node_paths) < 10, f"Suspiciously many inserted node paths for variant {variant}"

        did_add = False

        # If there are more than one path, some of them may be invalid (just by chance same sequence)
        # We only accespt paths where there is an edge from before the path to after the path
        possible_inserted_node_paths_orig = possible_inserted_node_paths.copy()
        if len(possible_inserted_node_paths) > 1:
            correct_paths = []
            for possible_path in possible_inserted_node_paths:
                in_nodes = self.mutable_graph.get_nodes_before(possible_path[0])
                out_nodes = self.mutable_graph.get_edges(possible_path[-1])

                if len([(from_node, to_node) for from_node, to_node in product(in_nodes, out_nodes) if to_node in self.graph.get_edges(from_node)]) > 0:

                    if variant.type == "DELETION":
                        # If deletion, the last position in the last node on the path must be on the linear reference and match the length of the deleted sequence
                        deletion_end_pos = variant.position + len(inserted_sequence)
                        if deletion_end_pos != self.graph.get_chromosome_ref_offset_at_node(variant.chromosome, possible_path[-1]) + self.graph.get_node_size(possible_path[-1]):
                            logging.debug("Ignoring deletion path %s because ref pos at end is not correct" % possible_path)
                            logging.debug(f"Deletion ends at position {deletion_end_pos}. Variant: {variant}")
                            continue

                    correct_paths.append(possible_path)

            possible_inserted_node_paths = correct_paths

        if len(possible_inserted_node_paths) == 0:
            logging.error("Did not find any valid inserted node paths for variant %s" % variant)
            logging.error("Possible paths are: %s" % possible_inserted_node_paths)
            logging.error("Original paths are: %s" % possible_inserted_node_paths_orig)
            raise VariantNotFoundException("Could not parse variatn")

        for inserted_nodes in possible_inserted_node_paths:

            # Find edges going from any node that goes into the inserted nodes and that ends at any node going out from the end of the inserted nodes
            # These edges are bypassing the inserted nodes
            in_nodes = self.mutable_graph.get_nodes_before(inserted_nodes[0])
            out_nodes = self.mutable_graph.get_edges(inserted_nodes[-1])
            #logging.info(" Out nodes: %s" % out_nodes)
            # This rule cannot be used since there might be an insertion right after a deletion that makes to_node not to be in edges from from_node
            # bypass_edges = [(from_node, to_node) for from_node, to_node in product(in_nodes, out_nodes) if to_node in self.mutable_graph.get_edges(from_node)]

            bypass_edges = set([(from_node, to_node) for from_node, to_node in product(in_nodes, out_nodes)])
            if len(bypass_edges) == 0:
                #logging.warning("Found no edges bypassing the inserted nodes %s. Nodes before are %s, nodes after are %s" % (inserted_nodes, in_nodes, out_nodes))
                #logging.warning("Variant is %s" % variant)
                continue

            #logging.info("Adding bypass edges: " + str(bypass_edges))
            self._add_new_dummy_node_for_edges(bypass_edges)
            did_add = True

        if not did_add:
            logging.warning("Not able to process variant %s" % variant)
            logging.warning("Possible inserted nodes in variant are %s" % possible_inserted_node_paths)
            raise VariantNotFoundException("No paths worked")

    def _add_new_dummy_node_for_edges(self, edges):
        dummy_node = self.current_new_node_id
        self.mutable_graph.add_node(dummy_node)

        # Remove all edges
        for from_node, to_node in edges:
            if to_node in self.mutable_graph.get_edges(from_node):
                self.mutable_graph.remove_edge(from_node, to_node)
                self._old_edges_to_new_node_mapping[(from_node, to_node)] = dummy_node

        # For all unique from_nodes add an edge to the dummy node
        for from_node in set((from_node for from_node, to_node in edges)):
            self.mutable_graph.add_edge(from_node, dummy_node)

        # For all unique to_nodes, add an edge from dummy node to to node
        for to_node in set((to_node for from_node, to_node in edges)):
            #print("Adding node from dummy node %d to node %d" % (dummy_node, to_node))
            self.mutable_graph.add_edge(dummy_node, to_node)

        self.current_new_node_id += 1




from setuptools import setup
from distutils.core import setup
from Cython.Build import cythonize
from Cython.Distutils import build_ext
import numpy as np

setup(name='obgraph',
      version='0.0.35',
      description='obgraph',
      url='http://github.com/ivargr/obgraph',
      author='Ivar Grytten',
      author_email='',
      license='MIT',
      packages=["obgraph"],
      zip_safe=False,
      install_requires=['cython', 'numpy', 'tqdm', 'pathos', 'shared_memory_wrapper>=0.0.18', 'bionumpy'],
      classifiers=[
            'Programming Language :: Python :: 3'
      ],
      entry_points={
            'console_scripts': ['obgraph=obgraph.command_line_interface:main']
      },
      cmdclass = {"build_ext": build_ext},
      ext_modules = cythonize(["obgraph/cython_traversing.pyx"]),
      include_dirs=np.get_include(),
)

"""

rm -rf dist
python3 setup.py sdist
twine upload --skip-existing dist/*


rm -rf dist
python3 setup.py sdist bdist_wheel
auditwheel repair --plat manylinux_2_17_x86_64 dist/obgraph-*-cp38-cp38-linux_x86_64.whl
rm dist/*.whl
mv wheelhouse/* dist
python3 -m twine upload --repository pypi dist/*

"""


from .variants import VcfVariants
import logging
from collections import defaultdict
from .mutable_graph import MutableGraph
from .graph import Graph
from .dummy_node_adder import DummyNodeAdder
import numpy as np

class EmptyNodeException(Exception):
    pass

class GraphConstructor:
    def __init__(self, reference_sequence, variants: VcfVariants):
        self.reference_sequence = reference_sequence
        self.variants = variants
        self.breakpoints = []
        self.make_sorted_breakpoints()

        self._ref_pos_to_node_after = defaultdict(list)
        self._node_to_ref_pos_after = defaultdict(list)
        self._node_to_ref_pos_before = defaultdict(list)
        self._ref_pos_to_node_start = defaultdict(list)

        self._node_ids = []
        self._node_sizes = []
        self._node_sequences = []
        self._reference_nodes = []
        self._current_node_id = 1
        self._edges_from = []
        self._edges_to = []

        self._edges_from_ref_pos_to_node = []
        self._deletions = defaultdict(list)  # A lookup from a ref pos to another ref pos representing a deltion
        self._edges_added = set()

        self._mutable_graph = MutableGraph()
        self.make_nodes()
        self.make_edges()
        self._graph = None
        self._graph_with_dummy_nodes = None
        logging.info("Making a graph")
        self._graph = self.get_graph()
        logging.info("Adding dummy nodes for indels")
        self.add_dummy_nodes()

    def get_graph(self):
        return Graph.from_mutable_graph(self._mutable_graph)

    def get_graph_with_dummy_nodes(self):
        return self._graph_with_dummy_nodes

    def add_dummy_nodes(self):
        dummy_node_adder = DummyNodeAdder(self._graph, self.variants)
        new_graph = dummy_node_adder.create_new_graph_with_dummy_nodes(self._mutable_graph)
        self._graph_with_dummy_nodes = new_graph

    def _make_edge(self, from_node, to_node, ref_pos_before_to_node):
        if (from_node, to_node) in self._edges_added:
            return

        assert to_node != from_node, "Trying to make edge from same node to same node %d to %d" % (from_node, to_node)

        self._edges_from.append(from_node)
        self._edges_to.append(to_node)
        self._edges_added.add((from_node, to_node))
        self._mutable_graph.add_edge(from_node, to_node)

        # Check if there is a deletion going from start of to_node, if so we should also have an edge past to node
        # Check if ref_pos_after comes in at a deletion, if so also add edges to after deletion
        for next_ref_pos in self._deletions[ref_pos_before_to_node+1]:
            for node_after_deletion in self._ref_pos_to_node_after[next_ref_pos - 1]:
                #logging.info("    Adding deletion edge from %d to %d" % (from_node, node_after_deletion))
                self._make_edge(from_node, node_after_deletion, next_ref_pos-1)


    def _make_node(self, ref_position_before_node, ref_position_after_node, sequence, is_ref_node=False, is_deletion=False):
        if sequence == "":
            logging.error("Ref pos before node: %d" % ref_position_before_node)
            raise EmptyNodeException("Empty sequence for node")

        self._ref_pos_to_node_after[ref_position_before_node].append(self._current_node_id)
        self._ref_pos_to_node_start[ref_position_before_node+1].append(self._current_node_id)
        #logging.info("         Addding node %d to ref_pos_to_node_after for ref pos %d" % (self._current_node_id, ref_position_before_node))
        self._node_to_ref_pos_after[self._current_node_id].append(ref_position_after_node)
        self._node_to_ref_pos_before[self._current_node_id].append(ref_position_before_node)
        node_id = self._current_node_id

        self._mutable_graph.add_node(node_id, sequence, is_ref_node=is_ref_node)

        self._current_node_id += 1
        return node_id

    def make_nodes(self):
        logging.info("Making nodes")
        prev_ref_node_end = -1

        self._mutable_graph.chromosome_start_nodes = {}

        i = 0
        for breakpoint_position, variant in self.breakpoints:


            i += 1

            if i % 100000 == 0:
                logging.info("%d/%d breakpoints processed" % (i, len(self.breakpoints)))

            #logging.info("At breakpoint %s, %s" % (breakpoint_position, variant))
            # Always make a ref variant from prev_ref_node_end to this breakpoints
            if breakpoint_position > prev_ref_node_end:
                #logging.info("   Making a reference node %d. Pos before/after: %d/%d." % (
                #self._current_node_id, prev_ref_node_end, breakpoint_position + 1))
                try:
                    sequence_between_breakpoints = self.reference_sequence[prev_ref_node_end+1:breakpoint_position+1]
                    if sequence_between_breakpoints == "":
                        logging.error("Empty sequence on ref node. Prev ref node ended at %d, breakpoint position: %s" % (prev_ref_node_end, breakpoint_position))
                        logging.error("Is your reference genome matching the vcf?")
                        assert False

                    prev_ref_node_id = self._make_node(prev_ref_node_end, breakpoint_position+1,
                                                       sequence_between_breakpoints, is_ref_node=True)

                except EmptyNodeException:
                    logging.error("Tried making a node between pos %d and %d" % (prev_ref_node_end, breakpoint_position+1))
                    logging.info("Variant is %s" % variant)
                    raise

                prev_ref_node_end = breakpoint_position
                #logging.info("    Setting prev ref node end to %d" % prev_ref_node_end)

            # If breakpoint is a new variant, make a variant node
            if variant is not None:
                if variant.chromosome not in self._mutable_graph.chromosome_start_nodes:
                    self._mutable_graph.chromosome_start_nodes[variant.chromosome] = prev_ref_node_id
                    logging.info(
                        f"Setting chromosome start node of chromosome {variant.chromosome} to be {prev_ref_node_id}")

                if variant.type != "DELETION":
                    #logging.info("   Making a variant node %d. Pos before/after: %d/%d. Variant node sequence: %s" % (
                    #self._current_node_id, variant.get_reference_position_before_variant(),
                    #variant.get_reference_position_after_variant(), variant.get_variant_sequence()))
                    self._make_node(variant.get_reference_position_before_variant(), variant.get_reference_position_after_variant(), variant.get_variant_sequence())
                else:
                    #logging.info("Adding deletion info from ref pos %d to ref pos %d" % (variant.get_reference_position_before_variant()+1, variant.get_reference_position_after_variant()))
                    self._deletions[variant.get_reference_position_before_variant()+1].append(variant.get_reference_position_after_variant())
                    #self._node_to_ref_pos_after[prev_ref_node_id].append(variant.get_reference_position_after_variant())

        # Add one reference node for the rest of the reference sequence that is left
        self._make_node(prev_ref_node_end, len(self.reference_sequence), self.reference_sequence[prev_ref_node_end+1:len(self.reference_sequence)], is_ref_node=True)

    def make_edges(self):
        #print("Ref pos to node after: %s" % self._ref_pos_to_node_after)
        #print("Ref pos to node start: %s" % self._ref_pos_to_node_start)
        #print("Node to ref pos after: %s" % self._node_to_ref_pos_after)
        #print("Node to ref pos before: %s" % self._node_to_ref_pos_before)
        #logging.info("Adding edges")
        i = 0
        # we want an edge from the node to nodes that come after this node
        # we find nodes coming after by looking at which ref position this node goes to
        # then we find all nodes where the ref position before that ref pos goes to the node
        for from_node, ref_positions_after in self._node_to_ref_pos_after.items():
            if i % 500000 == 0:
                logging.info("%d edge nodes processed" % i)
            i += 1
            for ref_pos_after in ref_positions_after:
                # Find all nodes where the ref pos before ref_pos_after goes to the node
                # ignore nodes with same ref pos before annd after as this node
                # such nodes are parallel nodes to this node and we don't want to connect them
                for to_node in self._ref_pos_to_node_after[ref_pos_after-1]:
                    if to_node == from_node:
                        continue
                    if self._node_to_ref_pos_after[from_node] == self._node_to_ref_pos_after[to_node] \
                            and self._node_to_ref_pos_before[from_node] == self._node_to_ref_pos_before[to_node]:
                        #print("Skipping edge from %d to %d" % (from_node, to_node))
                        continue
                    #print("Adding edge from %d to %d" % (from_node, to_node))
                    self._make_edge(from_node, to_node, ref_pos_after-1)


    def make_sorted_breakpoints(self):
        # Assume variants are already sorted
        logging.info("Traversing variants")
        for i, variant in enumerate(self.variants):
            if i % 10000 == 0:
                logging.info("%d variants processed" % i)

            # Breakpoints are last base pair in a reference node
            self.breakpoints.append((variant.get_reference_position_before_variant(), variant))
            self.breakpoints.append((variant.get_reference_position_after_variant()-1, None))

        logging.info("Sorting breakpoints")
        self.breakpoints = sorted(self.breakpoints, key=lambda b: b[0])
        logging.info("Done sorting breakpoints")

        #logging.info("Breakpoints: %s" % self.breakpoints)



import pytest
import time
import numpy as np
import bionumpy as bnp


@pytest.mark.skip
def test():
    t = time.perf_counter()
    file = bnp.open("variants.vcf", buffer_type=bnp.PhasedVCFMatrixBuffer)
    n = 0
    chunk = file.read_chunk()
    genotypes = chunk.genotypes.raw()
    print(np.unique(genotypes))
    n += chunk.genotypes.shape[0]
    print(n)
    print(time.perf_counter()-t)

test()


import numpy as np
import logging


class CoordinateConverter:
    def __init__(self, chromosome_offsets):
        self.chromosome_offsets = chromosome_offsets

    @classmethod
    def from_graph(cls, graph):
        chromosome_start_offsets = graph.node_to_ref_offset[graph.chromosome_start_nodes]
        return cls(chromosome_start_offsets)

    def convert_chromosome_offset(self, chromosome, offset):
        if chromosome == "X":
            chromosome = 23
        elif chromosome == "Y":
            chromosome = 24
        id = int(chromosome) - 1
        if id >= len(self.chromosome_offsets):
            return None

        return int(self.chromosome_offsets[id] + offset)

    def convert_offset_to_chromosome_and_offset(self, offset):
        index = np.searchsorted(self.chromosome_offsets, offset) - 1
        chromosome = str(index + 1)
        offset = offset - self.chromosome_offsets[index]

        return chromosome, int(offset)

    @classmethod
    def from_file(cls, file_name):
        data = np.load(file_name)
        return cls(data)

    def to_file(self, file_name):
        np.save(file_name, self.chromosome_offsets)